"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CircomExpressionVisitor = void 0;
const utils_1 = require("./utils");
const generated_1 = require("../generated");
const ParserError_1 = require("../errors/ParserError");
// TODO make operations modulo Q
class CircomExpressionVisitor extends generated_1.CircomVisitor {
    allowId;
    variablesContext;
    constructor(allowId, variablesContext = {}) {
        super();
        this.allowId = allowId;
        this.variablesContext = variablesContext;
    }
    visitExpression = (ctx) => {
        const expressionValue = this.visit(ctx);
        if (!(0, utils_1.validateBigInt)(expressionValue)) {
            throw new ParserError_1.ParserError({
                message: "Expression value must be of type bigint or bigint array",
                line: ctx.start.line,
                column: ctx.start.column,
            });
        }
        return expressionValue;
    };
    visitPrimaryExpression = (ctx) => {
        const primary = ctx.primary();
        if ((primary.identifier() && !this.allowId) || primary.args()) {
            throw new ParserError_1.ParserError({
                message: "Identifier usage is not allowed within the main component's parameters",
                line: primary.start.line,
                column: primary.start.column,
            });
        }
        else if (primary.identifier() && this.allowId) {
            const id = primary.identifier().ID().getText();
            if (!(id in this.variablesContext)) {
                throw new ParserError_1.ParserError({
                    message: `Unresolvable identifier ${id}`,
                    line: primary.start.line,
                    column: primary.start.column,
                });
            }
            const dimensions = (0, utils_1.resolveDimensions)(primary.identifier().arrayDimension_list(), this.variablesContext);
            let identifierValue = this.variablesContext[id].value;
            for (let i = 0; i < dimensions.length; i++) {
                if (Array.isArray(identifierValue)) {
                    identifierValue = identifierValue[Number(dimensions[i])];
                }
            }
            if (!(0, utils_1.validateBigInt)(identifierValue)) {
                throw new ParserError_1.ParserError({
                    message: "Unexpected type for identifier value",
                    line: primary.start.line,
                    column: primary.start.column,
                });
            }
            return identifierValue;
        }
        else if (primary.NUMBER()) {
            return BigInt(primary.NUMBER().getText());
        }
        else if (primary.numSequence()) {
            const numSequence = [];
            primary
                .numSequence()
                .NUMBER_list()
                .forEach((number) => {
                numSequence.push(BigInt(number.getText()));
            });
            return numSequence;
        }
        else if (primary.expression()) {
            return this.visitExpression(primary.expression());
        }
        else if (primary.expressionList()) {
            const expressionsResult = [];
            primary
                .expressionList()
                .expression_list()
                .forEach((expression) => {
                expressionsResult.push(this.visitExpression(expression));
            });
            return expressionsResult;
        }
        else {
            throw new ParserError_1.ParserError({
                message: "Unsupported expression",
                line: primary.start.line,
                column: primary.start.column,
            });
        }
    };
    visitBinaryExpression = (ctx) => {
        const left = this.visitExpression(ctx.expression(0));
        const right = this.visitExpression(ctx.expression(1));
        if (typeof left !== "bigint" || typeof right !== "bigint") {
            throw new ParserError_1.ParserError({
                message: "Expected bigint operands in binary expression",
                line: ctx.start.line,
                column: ctx.start.column,
            });
        }
        switch (ctx._op.type) {
            case generated_1.CircomParser.MUL:
                return left * right;
            case generated_1.CircomParser.DIV:
                if (right === BigInt(0)) {
                    throw new ParserError_1.ParserError({
                        message: "Division by zero is not allowed",
                        line: ctx.start.line,
                        column: ctx.start.column,
                    });
                }
                return left / right;
            case generated_1.CircomParser.ADD:
                return left + right;
            case generated_1.CircomParser.SUB:
                return left - right;
            case generated_1.CircomParser.MOD:
                return left % right;
            case generated_1.CircomParser.POW:
                return left ** right;
            case generated_1.CircomParser.SHL:
                return left << right;
            case generated_1.CircomParser.SHR:
                return left >> right;
            case generated_1.CircomParser.BAND:
                return left & right;
            case generated_1.CircomParser.BXOR:
                return left ^ right;
            case generated_1.CircomParser.BOR:
                return left | right;
            case generated_1.CircomParser.EQ:
                return left === right ? 1n : 0n;
            case generated_1.CircomParser.NEQ:
                return left !== right ? 1n : 0n;
            case generated_1.CircomParser.LT:
                return left < right ? 1n : 0n;
            case generated_1.CircomParser.GT:
                return left > right ? 1n : 0n;
            case generated_1.CircomParser.LE:
                return left <= right ? 1n : 0n;
            case generated_1.CircomParser.GE:
                return left >= right ? 1n : 0n;
            case generated_1.CircomParser.AND:
                return left !== 0n && right !== 0n ? 1n : 0n;
            case generated_1.CircomParser.OR:
                return left !== 0n || right !== 0n ? 1n : 0n;
            default:
                throw new ParserError_1.ParserError({
                    message: "Unsupported binary operator",
                    line: ctx.start.line,
                    column: ctx.start.column,
                });
        }
    };
    visitUnaryExpression = (ctx) => {
        const expressionValue = this.visitExpression(ctx.expression());
        if (typeof expressionValue !== "bigint") {
            throw new ParserError_1.ParserError({
                message: "Expected bigint operand in unary expression",
                line: ctx.start.line,
                column: ctx.start.column,
            });
        }
        if (ctx._op.type === generated_1.CircomParser.NOT) {
            return expressionValue === 0n ? 1n : 0n;
        }
        else if (ctx._op.type === generated_1.CircomParser.BNOT) {
            return ~expressionValue;
        }
        else {
            throw new ParserError_1.ParserError({
                message: "Unsupported unary operator",
                line: ctx.start.line,
                column: ctx.start.column,
            });
        }
    };
    visitTernaryExpression = (ctx) => {
        const conditionValue = this.visitExpression(ctx.expression(0));
        if (typeof conditionValue !== "bigint") {
            throw new ParserError_1.ParserError({
                message: "Expected bigint conditional value in ternary expression",
                line: ctx.start.line,
                column: ctx.start.column,
            });
        }
        if (conditionValue) {
            return this.visitExpression(ctx.expression(1));
        }
        else {
            return this.visitExpression(ctx.expression(2));
        }
    };
    visitDotExpression = (ctx) => {
        throw new ParserError_1.ParserError({
            message: "Dot expressions are not allowed within the main component's parameters",
            line: ctx.start.line,
            column: ctx.start.column,
        });
    };
}
exports.CircomExpressionVisitor = CircomExpressionVisitor;
