"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseIdentifier = parseIdentifier;
exports.validateBigInt = validateBigInt;
exports.resolveDimensions = resolveDimensions;
exports.setValueToArrayElement = setValueToArrayElement;
exports.setZeroValueToArrayElements = setZeroValueToArrayElements;
exports.validateArrayDimensions = validateArrayDimensions;
exports.performAssignmentOperation = performAssignmentOperation;
exports.performPostfixOperation = performPostfixOperation;
const CircomExpressionVisitor_1 = require("./CircomExpressionVisitor");
const constants_1 = require("./constants");
const ParserError_1 = require("../errors/ParserError");
function parseIdentifier(identifier) {
    const inputDimension = [];
    identifier.arrayDimension_list().forEach((dimension) => {
        inputDimension.push(dimension.getText().slice(1, -1));
    });
    return {
        name: identifier.ID().getText(),
        dimension: inputDimension,
    };
}
function validateBigInt(value) {
    if (typeof value === "bigint") {
        return true;
    }
    else if (Array.isArray(value)) {
        return value.every((element) => validateBigInt(element));
    }
    return false;
}
// returns  array of resolved indexes as bigints
// e.g. variable[0][2][1] => [0, 2, 1]
function resolveDimensions(arrayDimentions, variablesContext) {
    const dimensions = [];
    arrayDimentions.forEach((dimension) => {
        if (dimension.expression()) {
            const expressionVisitor = new CircomExpressionVisitor_1.CircomExpressionVisitor(true, variablesContext);
            const expressionValue = expressionVisitor.visitExpression(dimension.expression());
            if (typeof expressionValue !== "bigint") {
                throw new ParserError_1.ParserError({
                    message: `Dimension value type is not allowed to be ${typeof expressionValue}`,
                    line: dimension.start.line,
                    column: dimension.start.column,
                });
            }
            dimensions.push(Number(expressionValue));
        }
    });
    return dimensions;
}
function setValueToArrayElement(array, indices, value) {
    let current = array ? array : [];
    if (Array.isArray(current)) {
        for (let i = 0; i < indices.length; i++) {
            const index = indices[i];
            if (i === indices.length - 1) {
                current[index] = value;
            }
            else {
                if (!Array.isArray(current[index])) {
                    current[index] = [];
                }
                current = current[index];
            }
        }
    }
}
function setZeroValueToArrayElements(dimensions) {
    if (dimensions.length === 0) {
        return 0n;
    }
    const size = dimensions[0];
    const remainingDimensions = dimensions.slice(1);
    const array = new Array(size);
    for (let i = 0; i < size; i++) {
        array[i] = setZeroValueToArrayElements(remainingDimensions);
    }
    return array;
}
// level is a current depth in the recursive check
function validateArrayDimensions(array, dimensions, level = 0) {
    if (typeof array === "bigint" && !dimensions.length) {
        return true;
    }
    if (!Array.isArray(array)) {
        return false;
    }
    if (array.length !== dimensions[level]) {
        return false;
    }
    if (level < dimensions.length - 1) {
        for (let i = 0; i < array.length; i++) {
            if (!validateArrayDimensions(array[i], dimensions, level + 1)) {
                return false;
            }
        }
    }
    return true;
}
function performAssignmentOperation(assignmentOp, variables, variableId, leftValue, rightValue, ctxLocation) {
    // TODO make operations modulo Q
    switch (assignmentOp) {
        case constants_1.ASSIGNMENT_OPERATIONS.ADD: {
            variables[variableId].value = leftValue + rightValue;
            break;
        }
        case constants_1.ASSIGNMENT_OPERATIONS.SUB: {
            variables[variableId].value = leftValue - rightValue;
            break;
        }
        case constants_1.ASSIGNMENT_OPERATIONS.MUL: {
            variables[variableId].value = leftValue * rightValue;
            break;
        }
        case constants_1.ASSIGNMENT_OPERATIONS.DIV: {
            if (rightValue === 0n) {
                throw new ParserError_1.ParserError({
                    message: "Division by zero is not allowed",
                    line: ctxLocation.line,
                    column: ctxLocation.column,
                });
            }
            variables[variableId].value = leftValue / rightValue;
            break;
        }
        case constants_1.ASSIGNMENT_OPERATIONS.MOD: {
            variables[variableId].value = leftValue % rightValue;
            break;
        }
        case constants_1.ASSIGNMENT_OPERATIONS.EXP: {
            variables[variableId].value = leftValue ** rightValue;
            break;
        }
        case constants_1.ASSIGNMENT_OPERATIONS.LEFT_SHIFT: {
            variables[variableId].value = leftValue << rightValue;
            break;
        }
        case constants_1.ASSIGNMENT_OPERATIONS.RIGHT_SHIFT: {
            variables[variableId].value = leftValue >> rightValue;
            break;
        }
        case constants_1.ASSIGNMENT_OPERATIONS.BIT_AND: {
            variables[variableId].value = leftValue & rightValue;
            break;
        }
        case constants_1.ASSIGNMENT_OPERATIONS.BIT_OR: {
            variables[variableId].value = leftValue | rightValue;
            break;
        }
        case constants_1.ASSIGNMENT_OPERATIONS.BIT_XOR: {
            variables[variableId].value = leftValue ^ rightValue;
            break;
        }
        default: {
            throw new ParserError_1.ParserError({
                message: "Unsupported assignment operation",
                line: ctxLocation.line,
                column: ctxLocation.column,
            });
        }
    }
}
function performPostfixOperation(postfixOp, variables, variableId, ctxLocation) {
    if (typeof variables[variableId].value !== "bigint") {
        throw new ParserError_1.ParserError({
            message: "Expected bigint operands in postfix operation",
            line: ctxLocation.line,
            column: ctxLocation.column,
        });
    }
    if (postfixOp === constants_1.POSTFIX_OPERATIONS.INCR) {
        variables[variableId].value += 1n;
    }
    else if (postfixOp === constants_1.POSTFIX_OPERATIONS.DECR) {
        variables[variableId].value -= 1n;
    }
    else {
        throw new ParserError_1.ParserError({
            message: `Unsupported postfix operator ${postfixOp}`,
            line: ctxLocation.line,
            column: ctxLocation.column,
        });
    }
}
